# -*- coding: utf-8 -*-

import shlex
import subprocess
from unittest import TestCase

import pandas
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class Testround_index(TestCase):
    def setUp(self):
        dr = pandas.date_range("2000-01-01", periods=2, freq="D")

        ts = pandas.Series([4.5, 4.6], index=dr)

        self.round_index_direct = pandas.DataFrame(ts, columns=["Value"])
        self.round_index_direct.index.name = "Datetime"
        self.round_index_direct = tsutils.memory_optimize(self.round_index_direct)

        self.round_index_multiple_direct = pandas.DataFrame(ts, columns=["Value"])
        self.round_index_multiple_direct = pandas.concat(
            [self.round_index_multiple_direct, pandas.Series(ts, name="Value_r")],
            axis="columns",
        )
        self.round_index_multiple_direct.index.name = "Datetime"
        self.round_index_multiple_direct = tsutils.memory_optimize(
            self.round_index_multiple_direct
        )

        self.round_index_cli = b"""Datetime,Value
2000-01-01,4.5
2000-01-02,4.6
"""

        self.round_index_multiple_cli = b"""Datetime,Value,Value_r
2000-01-01,4.5,4.5
2000-01-02,4.6,4.6
"""

        self.round_index_tsstep_2_daily_cli = b"""Datetime,Value,Value1
2000-01-01,4.5,45.6
2000-01-03,4.7,34.2
2000-01-05,4.5,7.2
"""
        self.round_index_tsstep_2_daily = pandas.DataFrame(
            [[4.5, 45.6], [4.7, 34.2], [4.5, 7.2]],
            columns=["Value", "Value1"],
            index=pandas.DatetimeIndex(["2000-01-01", "2000-01-03", "2000-01-05"]),
        )
        self.round_index_tsstep_2_daily = tsutils.memory_optimize(
            self.round_index_tsstep_2_daily
        )

        self.round_index_tsstep_2_daily.index.name = "Datetime"

        self.round_index_blanks = b"""Datetime,Value_mean,Unnamed: 2_mean,Unnamed: 3_mean,Unnamed: 4_mean,Unnamed: 5_mean,Unnamed: 6_mean,Unnamed: 7_mean,Unnamed: 8_mean,Unnamed: 9_mean
2000-01-01,2.46667,,,,,,,,
2000-01-02,3.4,,,,,,,,
"""

    def test_round_index_direct(self):
        """Test round_index API for single column - daily."""
        out = tstoolbox.read("tests/data_simple.csv", round_index="D")
        assert_frame_equal(out, self.round_index_direct)

    def test_round_index_mulitple_direct(self):
        """Test round_index API for multiple columns - daily."""
        out = tstoolbox.read(
            "tests/data_simple.csv tests/data_simple.csv",
            append="columns",
            round_index="D",
        )
        assert_frame_equal(out, self.round_index_multiple_direct)

    def test_round_index_bi_monthly(self):
        """Test round_index API for bi monthly time series."""
        out = tstoolbox.read("tests/data_bi_daily.csv", round_index="D")
        assert_frame_equal(out, self.round_index_tsstep_2_daily)

    def test_round_index_cli(self):
        """Test round_index CLI for single column - daily."""
        args = 'tstoolbox read --round_index="D" tests/data_simple.csv'
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.round_index_cli)

    def test_round_index_multiple_cli(self):
        """Test round_index CLI for multiple columns - daily."""
        args = 'tstoolbox read --round_index="D" tests/data_simple.csv tests/data_simple.csv'
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.round_index_multiple_cli)

    def test_round_index_bi_monthly_cli(self):
        """Test round_index CLI for bi monthly time series."""
        args = 'tstoolbox read --round_index="D" tests/data_bi_daily.csv'
        args = shlex.split(args)
        out = subprocess.Popen(args, stdout=subprocess.PIPE).communicate()
        self.assertEqual(out[0], self.round_index_tsstep_2_daily_cli)



# -*- coding: utf-8 -*-

from unittest import TestCase

import pandas as pd
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class TestDescribe(TestCase):
    def setUp(self):
        self.data_0_to_1 = tstoolbox.read("tests/data_sunspot_normalized_0_to_1.csv")
        self.data_0_to_1.columns = ["Area::minmax"]
        self.data_10_to_20 = tstoolbox.read(
            "tests/data_sunspot_normalized_10_to_20.csv"
        )
        self.data_10_to_20.columns = ["Area::minmax"]
        data_zscore = tstoolbox.read("tests/data_sunspot_normalized_zscore.csv")
        self.data_zscore = pd.DataFrame(data_zscore.values, index=data_zscore.index)
        self.data_zscore.columns = ["Area::zscore"]
        self.data_zscore = tsutils.memory_optimize(self.data_zscore)
        self.data_pct_rank = tstoolbox.read(
            "tests/data_sunspot_normalized_pct_rank.csv"
        )
        self.data_pct_rank.columns = ["Area::pct_rank"]
        self.data_pct_rank = tsutils.memory_optimize(self.data_pct_rank)

    def test_normalize_0_to_1(self):
        """Test the normalization API function from 0 to 1."""
        out = tstoolbox.normalization(input_ts="tests/data_sunspot.csv")
        assert_frame_equal(out, self.data_0_to_1)

    def test_normalize_10_to_20(self):
        """Test the normalization API function from 10 to 20."""
        out = tstoolbox.normalization(
            min_limit=10, max_limit=20, input_ts="tests/data_sunspot.csv"
        )
        assert_frame_equal(out, self.data_10_to_20)

    # The following test gave a very strange error where self.data_zscore
    # was and ExtensionArray
    # def test_normalize(self):
    #     """Test the normalization API function using the zscore method."""
    #     out = tstoolbox.normalization(mode="zscore",
    #                                   input_ts="tests/data_sunspot.csv")
    #     assert_frame_equal(
    #         out,
    #         self.data_zscore,
    #         check_exact=False,
    #         check_column_type=False,
    #         atol=1e-4
    #     )

    def test_pct_rank(self):
        """Test the normalization API function using the pct_rank method."""
        out = tstoolbox.normalization(
            mode="pct_rank", input_ts="tests/data_sunspot.csv"
        )
        assert_frame_equal(out, self.data_pct_rank)



# -*- coding: utf-8 -*-
import os

import xlsxwriter


def write_stats(filename, data, **kwargs):
    """
    Write statistics to an Excel file.

    This function writes to an Excel file FILENAME the statistics
    provided in each of the dictionaries contained in DATA. The first
    2 arguments must be the inputs as described below followed by
    keyword arguments in the format of OPTION = VALUE. Each statistic
    will be labeled according to the name under which it is stored in
    the DATA data structure, e.g. data.bias will be labeled as "bias".

    INPUTS:
    filename     : name for statistics Excel file
    data         : a dictionary containing the statistics
    data['stat'] : statistics, e.g. data.bias for Bias.

    OUTPUTS:
        None.

    LIST OF OPTIONS:
      A title description for each dictionary (TITLE) can be
    optionally provided as well as an overwrite option if the file
    name currently exists.

    title = title : title descriptor data set, e.g. 'Expt. 01.0'
    overwrite = boolean : true/false flag to overwrite Excel file

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 10, 2016
    """
    option = get_write_stats_options(**kwargs)

    # Check for existence of file
    if os.path.isfile(filename):
        if option["overwrite"]:
            os.remove(filename)
        else:
            ValueError("File already exists: " + filename)

    # Write title information to file
    workbook = xlsxwriter.Workbook(filename)
    worksheet = workbook.add_worksheet()

    # Write descriptive title
    if len(option["title"]) > 0:
        worksheet.write(1, 0, option["title"])
    else:
        worksheet.write(1, 0, "Skill Metrics")

    # Determine number of elements in the dictionary lists and write
    # appropriate header
    worksheet.write(3, 0, "Skill Metric")
    ncell = len(list(data.items())[0]) - 1
    for i in range(ncell):
        worksheet.write(3, i + 1, "Case " + str(i + 1))

    # Write data of all the fields
    row = 4
    col = 0
    for key, value in data.items():
        worksheet.write(row, col, key)
        for i, v in enumerate([value]):
            worksheet.write(row, col + 1 + i, v)
        row += 1

    workbook.close()


def get_write_stats_options(**kwargs):
    """
    Get optional arguments for write_stats function.

    Retrieves the keywords supplied to the WRITE_STATS function (**KWARGS),
    and returns the values in a OPTION dicitonary. Default values are
    assigned to selected optional arguments. The function will terminate
    with an error if an unrecognized optional argument is supplied.

    INPUTS:
    **kwargs : keyword argument list

    OUTPUTS:
    option : data structure containing option values.
    option['title']     : title descriptor for data set.
    option['overwrite'] : boolean to overwrite Excel file.

    LIST OF OPTIONS:
      A title description for each dataset TITLE can be optionally
    provided as well as an overwrite option if the file name currently
    exists.

    title = title : title descriptor for each data set in data, e.g.
                   'Expt. 01.0'
    overwrite = boolean : true/false flag to overwrite Excel file

    Author: Peter A. Rochford
    Acorn Science & Innovation
    prochford@acornsi.com

    Created on Dec 10, 2016

    @author: rochfordp
    """
    from . import check_on_off

    nargin = len(kwargs)

    #  Set default parameters
    option = {}
    option["title"] = ""
    option["overwrite"] = False
    if nargin == 0:
        # No options requested, so return with only defaults
        return option

    # Load custom options, storing values in option data structure

    # Check for valid keys and values in dictionary
    for optname, optvalue in kwargs.items():
        optname = optname.lower()
        if optname not in option:
            raise ValueError("Unrecognized option: " + optname)
        # Replace option value with that from arguments
        option[optname] = optvalue

        # Check values for specific options
        if optname == "overwrite":
            option["overwrite"] = check_on_off(option["overwrite"])

    return option



# -*- coding: utf-8 -*-

import shlex
import subprocess
from unittest import TestCase

import pandas as pd
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils

from . import capture


class TestDate_slice(TestCase):
    def setUp(self):
        dr = pd.date_range("2011-01-01T12:00:00", periods=3, freq="H")
        self.date_slice = pd.DataFrame([2, 2, 2], index=dr, columns=["Value"])
        self.date_slice = tsutils.memory_optimize(self.date_slice)

        self.date_slice_cli = capture.capture(tsutils.printiso, self.date_slice)

    def test_date_slice(self):
        """Test date_slice API."""
        out = tstoolbox.date_slice(
            input_ts="tests/data_flat.csv",
            start_date="2011-01-01T12:00:00",
            end_date="2011-01-01T14:00:00",
        )
        assert_frame_equal(out, self.date_slice)

    def test_date_slice_cli(self):
        """Test date_slice CLI."""
        args = (
            "tstoolbox date_slice "
            '--input_ts="tests/data_flat.csv" '
            '--start_date="2011-01-01T12:00:00" '
            '--end_date="2011-01-01T14:00:00"'
        )
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        self.assertEqual(out, self.date_slice_cli)



# -*- coding: utf-8 -*-

from unittest import TestCase

import pandas
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class TestRead(TestCase):
    def setUp(self):
        dr = pandas.date_range("2000-01-01", periods=2, freq="D")

        ts = pandas.Series([4.5, 4.6], index=dr)

        self.read_direct = pandas.DataFrame(ts, columns=["Value"])
        self.read_direct.index.name = "Datetime"
        self.read_direct = tsutils.memory_optimize(self.read_direct)

        self.read_cli = b"""Datetime,Value
2000-01-01,4.5
2000-01-02,4.6
"""

        dr = pandas.date_range("2000-01-01", periods=5, freq="D")

        ts = pandas.Series([4.5, 4.6, 4.7, 4.8, 4.9], index=dr)

        self.read_direct_sparse = pandas.DataFrame(ts, columns=["Value"])
        self.read_direct_sparse.index.name = "Datetime"
        self.read_direct_sparse = tsutils.memory_optimize(self.read_direct_sparse)

        self.read_cli_sparse = b"""Datetime,Value
2000-01-01,4.5
2000-01-02,4.6
2000-01-03,4.7
2000-01-04,4.8
2000-01-05,4.9
"""

    def test_read_direct(self):
        """Test read API for single column - daily."""
        out = tstoolbox.read("tests/data_simple_extra_rows.csv", skiprows=2)
        assert_frame_equal(out, self.read_direct)

    def test_read_direct_sparse(self):
        """Test read API for single column - daily."""
        out = tstoolbox.read("tests/data_simple_extra_rows_sparse.csv", skiprows=[4, 6])
        assert_frame_equal(out, self.read_direct_sparse)



# -*- coding: utf-8 -*-
import numpy as np


def get_taylor_diagram_options(*args, **kwargs):
    """
    Get optional arguments for taylor_diagram function.

    Retrieves the optional arguments supplied to the TAYLOR_DIAGRAM
    function as a variable-length input argument list (*ARGS), and
    returns the values in an OPTION dictionary. Default values are
    assigned to selected optional arguments. The function will terminate
    with an error if an unrecognized optional argument is supplied.

    INPUTS:
    *kwargs : variable-length keyword argument list. The keywords by
              definition are dictionaries with keys that must correspond to
              one choices given in OUTPUTS below.

    OUTPUTS:
    option : dictionary containing option values. (Refer to
             display_taylor_diagram_options function for more information.)
    option['alpha']           : blending of symbol face color (0.0
                                transparent through 1.0 opaque).
                                (Default: 1.0)
    option['axismax']         : maximum for the radial contours
    option['checkstats']      : Check input statistics satisfy Taylor
                                relationship (Default : 'off')
    option['colcor']   : color for correlation coefficient labels
                         (Default: blue)
    option['colobs']   : color for observation labels (Default : magenta)
    option['colrms']   : color for RMS labels (Default : medium green)
    option['colstd']   : color for STD labels (Default : black)
    option['colormap']        : 'on'/'off' switch to map color shading of
                                 markers to colormap ('on') or min to max range
                                 of RMSDz values ('off'). Set to same value as
                                 option['nonrmsdz'].
    option['markercolor']     : single color to use for all markers
                                (Default: red)
    option['markerdisplayed'] : markers to use for individual experiments
    option['markerlabel']     : name of the experiment to use for marker
    option['markerlabelcolor']: marker label color (Default : 'k')
    option['markerlegend']    : 'on'/'off' switch to display marker legend
                                (Default 'off')
    option['markerobs'  ]     : marker to use for x-axis indicating observed
                                STD. A choice of 'none' will suppress
                                appearance of marker. (Default 'none')
    option['markersize']      : marker size (Default 10)

    option['nonrmsdz']      : 'on'/'off' switch indicating values in RMSDz
                               do not correspond to total RMS Differences.
                               (Default 'off')
    option['numberpanels']  : Number of panels to display
                              = 1 for positive correlations
                              = 2 for positive and negative correlations
                             (Default value depends on correlations (cors))

    option['overlay']       : 'on'/'off' switch to overlay current
                                statistics on Taylor diagram (Default 'off')
                                Only markers will be displayed.
    option['rincrms']       : axis tick increment for RMS values
    option['rincstd']       : axis tick increment for STD values
    option['rmslabelformat'] : string format for RMS contour labels,
                               e.g. '0:.2f'.
                               (Default '0', format as specified by str
                               function)

    option['showlabelscor'] : show correlation coefficient labels
                              (Default: 'on')
    option['showlabelsrms'] : show RMS labels (Default: 'on')
    option['showlabelsstd'] : show STD labels (Default: 'on')

    option['stylecor']      : line style for correlation coefficient grid
                              lines (Default: dash-dot '-.')
    option['styleobs']      : line style for observation grid line. A choice of
                              empty string '' will suppress appearance of the
                              grid line (Default: '')
    option['stylerms']      : line style for RMS grid lines
                              (Default: dash '--')
    option['stylestd']      : line style for STD grid lines
                              (Default: dotted ':')

    option['tickcor'][panel]: tick values for correlation coefficients for
                              two types of panels
    option['tickrms']       : RMS values to plot grid circles from
                              observation point
    option['tickstd']       : STD values to plot grid circles from
                              origin
    option['tickrmsangle']  : tick RMS angle (Default: 135 degrees)
    option['titleColorBar'] : title for the colorbar
    option['titlecor']      : show correlation coefficient axis label
                              (Default: 'on')
    option['titleobs']      : label for observation point (Default: '')
    option['titlerms']      : show RMS axis label (Default: 'on')
    option['titlestd']      : show STD axis label (Default: 'on')

    option['widthcor']      : linewidth for correlation coefficient grid
                              lines (Default: .8)
    option['widthobs']      : linewidth for observation grid line (Default: .8)
    option['widthrms']      : linewidth for RMS grid lines (Default: .8)
    option['widthstd']      : linewidth for STD grid lines (Default: .8)

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Nov 25, 2016
    Revised on Apr  22, 2017
    """
    from . import check_on_off

    cors = args[0]
    nargin = len(kwargs)

    # Set default parameters for all options
    option = {}
    option["alpha"] = 1.0
    option["axismax"] = 0.0
    option["checkstats"] = "off"

    option["colcor"] = (0, 0, 1)  # blue
    option["colobs"] = "m"  # magenta
    option["colrms"] = (0, 0.6, 0)  # medium green
    option["colstd"] = (0, 0, 0)  # black

    option["colormap"] = "off"

    option["markercolor"] = "r"
    option["markerdisplayed"] = "marker"
    option["markerlabel"] = ""
    option["markerlabelcolor"] = "k"
    option["markerlegend"] = "off"
    option["markerobs"] = "none"
    option["markersize"] = 10

    option["nonrmsdz"] = "off"
    negative = cors[np.where(cors < 0.0)]
    if len(negative) > 0:
        option["numberpanels"] = 2  # double panel
    else:
        option["numberpanels"] = 1  # single panel

    option["overlay"] = "off"
    option["rincrms"] = []
    option["rincstd"] = []
    option["rmslabelformat"] = "0"

    option["showlabelscor"] = "on"
    option["showlabelsrms"] = "on"
    option["showlabelsstd"] = "on"

    option["stylecor"] = "-."
    option["styleobs"] = ""
    option["stylerms"] = "--"
    option["stylestd"] = ":"

    # Note that "0" must be explicitly given or a scientific number is
    # stored
    tickval1 = [1, 0.99, 0.95, 0]
    middle = np.linspace(0.9, 0.1, 9)
    tickval1[3:3] = middle
    tickval2 = tickval1[:]
    values = np.linspace(-0.1, -0.9, 9)
    tickval2.extend(values)
    tickval2.extend([-0.95, -0.99, -1])
    option["tickcor"] = (tickval1, tickval2)  # store as tuple

    option["tickrms"] = []
    option["tickstd"] = []
    option["tickrmsangle"] = -1
    option["titlecolorbar"] = ""
    option["titlecor"] = "on"
    option["titleobs"] = ""
    option["titlerms"] = "on"
    option["titlestd"] = "on"

    option["widthcor"] = 1.0
    option["widthobs"] = 1.0
    option["widthrms"] = 1.0
    option["widthstd"] = 1.0

    if nargin == 0:
        # No options requested, so return with only defaults
        return option

    # Check for valid keys and values in dictionary
    for optname, optvalue in kwargs.items():
        optname = optname.lower()
        if optname not in option:
            raise ValueError("Unrecognized option: " + optname)
        # Replace option value with that from arguments
        if optname == "tickcor":
            list1 = option["tickcor"][0]
            list2 = option["tickcor"][1]
            if option["numberpanels"] == 1:
                list1 = optvalue
            else:
                list2 = optvalue
            option["tickcor"] = (list1, list2)
        else:
            option[optname] = optvalue

        # Check values for specific options
        if optname == "checkstats":
            option["checkstats"] = check_on_off(option["checkstats"])
        elif optname == "markerlabel":
            if not isinstance(optvalue, list):
                raise ValueError("Option value is not a list: " + str(optvalue))
            option["markerlabel"] = optvalue[1:]
        elif optname == "markerlegend":
            option["markerlegend"] = check_on_off(option["markerlegend"])
        elif optname == "overlay":
            option["overlay"] = check_on_off(option["overlay"])
        elif optname == "rmslabelformat":
            # Check for valid string format
            labelformat = "{" + optvalue + "}"
            try:
                labelformat.format(99.0)
            except ValueError:
                raise ValueError(
                    """
        *
        *   Invalid string format for rmslabelformat: {}
        *
        """.format(
                        optvalue
                    )
                )
        elif optname == "showlabelscor":
            option["showlabelscor"] = check_on_off(option["showlabelscor"])
        elif optname == "showlabelsrms":
            option["showlabelsrms"] = check_on_off(option["showlabelsrms"])
        elif optname == "showlabelsstd":
            option["showlabelsstd"] = check_on_off(option["showlabelsstd"])
        elif optname == "tickrms":
            option["tickrms"] = np.sort(optvalue)
            option["rincrms"] = (max(option["tickrms"]) - min(option["tickrms"])) / len(
                option["tickrms"]
            )
        elif optname == "tickstd":
            option["tickstd"] = np.sort(optvalue)
            option["rincstd"] = (max(option["tickstd"]) - min(option["tickstd"])) / len(
                option["tickstd"]
            )
        elif optname == "titlecor":
            option["titlecor"] = check_on_off(option["titlecor"])
        elif optname == "titlerms":
            option["titlerms"] = check_on_off(option["titlerms"])
        elif optname == "titlestd":
            option["titlestd"] = check_on_off(option["titlestd"])

    option["colormap"] = option["nonrmsdz"]

    return option



# -*- coding: utf-8 -*-

from unittest import TestCase

import pint_pandas
import pytest
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox


class TestConvertUnits(TestCase):
    @staticmethod
    def test_convert_units():
        a = tstoolbox.read("tests/data_gainesville_daily_precip.csv", target_units="in")
        b = tstoolbox.equation(
            "x1/25.4", input_ts="tests/data_gainesville_daily_precip.csv"
        )
        b.columns = ["ADaymet-prcp:in"]
        assert_frame_equal(a, b, check_dtype=False)

        a = tstoolbox.read("tests/data_gainesville_daily_precip.csv", target_units="km")
        b = tstoolbox.equation(
            "x1/(1000*1000)", input_ts="tests/data_gainesville_daily_precip.csv"
        )
        b.columns = ["ADaymet-prcp:km"]
        assert_frame_equal(a, b, check_dtype=False)

        with pytest.raises(ValueError) as e_info:
            _ = tstoolbox.read(
                "tests/data_gainesville_daily_precip.csv", source_units="ft3/s"
            )
        assert r'The units specified by the "source_units" keyword and in the' in str(
            e_info.value
        )

        with pytest.raises(ValueError) as e_info:
            _ = tstoolbox.read(
                "tests/data_gainesville_daily_precip.csv", target_units="ft3/s"
            )



# -*- coding: utf-8 -*-
import numpy as np

from . import utils


def bias(predicted, reference):
    """
    Calculate the bias between PREDICTED and REFERENCE.

    B = mean(p) - mean(r)

    where p is the predicted values, and r is the reference values.
    Note that p & r must have the same number of values.

    Input:
    PREDICTED : predicted field
    REFERENCE : reference field

    Output:
    B : bias between predicted and reference

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 9, 2016
    """
    # Check that dimensions of predicted and reference fields match
    utils.check_arrays(predicted, reference)

    # Calculate bias in means
    b = np.mean(predicted) - np.mean(reference)

    return b



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings
from typing import List, Optional, Union

import mando
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter
from sklearn import linear_model
from sklearn.metrics import mean_squared_error, r2_score

from .. import tsutils

warnings.filterwarnings("ignore")

_FUNCS = {
    "ARD": linear_model.ARDRegression,
    "BayesianRidge": linear_model.BayesianRidge,
    "ElasticNetCV": linear_model.ElasticNetCV,
    "ElasticNet": linear_model.ElasticNet,
    "Huber": linear_model.HuberRegressor,
    "LarsCV": linear_model.LarsCV,
    "Lars": linear_model.Lars,
    "LassoCV": linear_model.LassoCV,
    "LassoLarsCV": linear_model.LassoLarsCV,
    "LassoLarsIC": linear_model.LassoLarsIC,
    "LassoLars": linear_model.LassoLars,
    "Lasso": linear_model.Lasso,
    "Linear": linear_model.LinearRegression,
    #    "LogisticCV": linear_model.LogisticRegressionCV,
    #    "Logistic": linear_model.LogisticRegression,
    #    "MultiTaskElasticNetCV": linear_model.MultiTaskElasticNetCV,
    #    "MultiTaskElasticNet": linear_model.MultiTaskElasticNet,
    #    "MultiTaskLassoCV": linear_model.MultiTaskLassoCV,
    #    "MultiTaskLasso": linear_model.MultiTaskLasso,
    "OrthogonalMatchingPursuitCV": linear_model.OrthogonalMatchingPursuitCV,
    "OrthogonalMatchingPursuit": linear_model.OrthogonalMatchingPursuit,
    #    "PassiveAggressive": linear_model.PassiveAggressiveClassifier,
    #    "Perceptron": linear_model.Perceptron,
    "RANSAC": linear_model.RANSACRegressor,
    "RidgeCV": linear_model.RidgeCV,
    "Ridge": linear_model.Ridge,
    "SGD": linear_model.SGDRegressor,
    "TheilSen": linear_model.TheilSenRegressor,
}


@mando.command("regression", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def regression_cli(
    method,
    x_train_cols,
    y_train_col,
    x_pred_cols=None,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    clean=False,
    round_index=None,
    skiprows=None,
    index_type="datetime",
    names=None,
    print_input=False,
    tablefmt="csv",
    por=False,
):
    """Regression of one or more time-series or indices to a time-series.

    If optional ``x_pred_cols`` is given will return a time-series of the ``y``
    predictions. Otherwise returns dictionary of equation and statistics about
    the regression fit.

    Parameters
    ----------
    method: str
        The method of regression.  The chosen method will use `x_train_cols` as the
        independent data and `y_pred_col` as the dependent data.

        ARD
            Requires lots of memory.

            Fit the weights of a regression model, using an ARD prior. The
            weights of the regression model are assumed to be in Gaussian
            distributions. Also estimate the parameters lambda (precisions of
            the distributions of the weights) and alpha (precision of the
            distribution of the noise). The estimation is done by an iterative
            procedures (Evidence Maximization)
        BayesianRidge
            Fit a Bayesian ridge model. See the Notes section for details on
            this implementation and the optimization of the regularization
            parameters lambda (precision of the weights) and alpha (precision
            of the noise).
        ElasticNetCV
            Elastic Net model with iterative fitting along a regularization
            path.
        ElasticNet
            Linear regression with combined L1 and L2 priors as regularizer.
        Huber
            Linear regression model that is robust to outliers.

            The Huber Regressor optimizes the squared loss for the samples
            where abs((y - X'w) / sigma) < epsilon and the absolute loss for
            the samples where abs((y - X'w) / sigma) > epsilon, where w and
            sigma are parameters to be optimized. The parameter sigma makes
            sure that if y is scaled up or down by a certain factor, one does
            not need to rescale epsilon to achieve the same robustness. Note
            that this does not take into account the fact that the different
            features of X may be of different scales.

            This makes sure that the loss function is not heavily influenced by
            the outliers while not completely ignoring their effect.
        LarsCV
            Cross-validated Least Angle Regression model.
        Lars
            Least Angle Regression model.
        LassoCV
            Lasso linear model with iterative fitting along a regularization
            path.
        LassoLarsCV
            Cross-validated Lasso, using the LARS algorithm.
        LassoLarsIC
            Lasso model fit with Lars using BIC or AIC for model selection.
        LassoLars
            Lasso model fit with Least Angle Regression a.k.a. Lars.  It is
            a Linear Model trained with an L1 prior as regularizer.
        Lasso
            Linear Model trained with L1 prior as regularizer (aka the Lasso).
        Linear
            LinearRegression fits a linear model with coefficients w = (w1, …,
            wp) to minimize the residual sum of squares between the observed
            targets in the dataset, and the targets predicted by the linear
            approximation.
        RANSAC
            RANSAC (RANdom SAmple Consensus) algorithm.  RANSAC is an iterative
            algorithm for the robust estimation of parameters from a subset of
            inliers from the complete data set.
        RidgeCV
            Ridge regression with built-in cross-validation.  By default, it
            performs Generalized Cross-Validation, which is a form of efficient
            Leave-One-Out cross-validation.
        Ridge
            This classifier first converts the target values into (-1, 1) and
            then treats the problem as a regression task (multi-output
            regression in the multiclass case).
        SGD
            Input must be scaled by removing mean and scaling to unit variance.
            Can use 'tstoolbox normalization ...' to scale the input.

            Linear model fitted by minimizing a regularized empirical loss with
            SGD.  SGD stands for Stochastic Gradient Descent: the gradient of
            the loss is estimated each sample at a time and the model is
            updated along the way with a decreasing strength schedule (aka
            learning rate).

            The regularizer is a penalty added to the loss function that
            shrinks model parameters towards the zero vector using either the
            squared euclidean norm L2 or the absolute norm L1 or a combination
            of both (Elastic Net). If the parameter update crosses the 0.0
            value because of the regularizer, the update is truncated to 0.0 to
            allow for learning sparse models and achieve online feature
            selection.
        TheilSen
            Theil-Sen Estimator: robust multivariate regression model.

            The algorithm calculates least square solutions on subsets with
            size n_subsamples of the samples in X. Any value of n_subsamples
            between the number of features and samples leads to an estimator
            with a compromise between robustness and efficiency. Since the
            number of least square solutions is “n_samples choose
            n_subsamples”, it can be extremely large and can therefore be
            limited with max_subpopulation. If this limit is reached, the
            subsets are chosen randomly. In a final step, the spatial median
            (or L1 median) is calculated of all least square solutions.
    x_train_cols: str or list
        List of column names/numbers that hold the ``x`` value datasets used to
        train the regression.  Perform a multiple regression if ``method``
        allows by giving several ``x_train_cols``.  To include the index in the
        regression use column 0 or the index name.
    y_train_col: str or list
        Column name or number of the ``y`` dataset used to train the
        regression.

        The ``y_train_col`` cannot be part of ``x_train_cols`` or
        ``x_pred_cols``.
    x_pred_cols : str or list
        [optional, if supplied will return a time-series of the ``y``
        prediction based on ``x_pred_cols``.]

        List of column names/numbers of ``x`` value datasets used to create the
        ``y`` prediction.  Needs to be the same number of columns as
        ``x_train_cols``.  Can be identical columns to ``x_train_cols``.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${clean}
    ${round_index}
    ${skiprows}
    ${index_type}
    ${names}
    ${print_input}
    ${tablefmt}
    ${por}
    """
    tsutils.printiso(
        regression(
            method,
            x_train_cols,
            y_train_col,
            x_pred_cols=x_pred_cols,
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            clean=clean,
            round_index=round_index,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            print_input=print_input,
            por=por,
        ),
        tablefmt=tablefmt,
        headers=[],
    )


# @tsutils.validator(
#    method=[str, ["domain", _FUNCS.keys()], None],
# )
@tsutils.transform_args(
    x_train_cols=tsutils.make_list,
    y_train_col=tsutils.make_list,
    x_pred_cols=tsutils.make_list,
)
@typic.al
@tsutils.copy_doc(regression_cli)
def regression(
    method,
    x_train_cols: List[Union[str, int]],
    y_train_col: List[Union[str, int]],
    x_pred_cols: Optional[List[Union[str, int]]] = None,
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    clean=False,
    round_index=None,
    skiprows=None,
    index_type="datetime",
    names=None,
    print_input=False,
    por=False,
):
    """Regression of data."""
    for to in y_train_col:
        for fro in x_train_cols:
            if to == fro:
                raise ValueError(
                    tsutils.error_wrapper(
                        f"""
You can't have columns in both "x_train_cols", and "y_train_col"
keywords.  Instead you have "{to}" in both.
"""
                    )
                )

    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        clean=clean,
        por=por,
    )

    if print_input is True:
        ntsd = tsd.copy()
    else:
        ntsd = tsd

    ntsd = tsutils.asbestfreq(ntsd)

    testfreqstr = ntsd.index.freqstr.lstrip("0123456789")
    if testfreqstr[0] == "A":
        ntsd[ntsd.index.name + "_"] = ntsd.index.year - ntsd.index[0].year
    elif testfreqstr[0] == "M":
        ntsd[ntsd.index.name + "_"] = (ntsd.index.year - ntsd.index[0].year) * 12 + (
            ntsd.index.month - ntsd.index[0].month
        )
    else:
        try:
            # In case ntsd.index.freqstr is a multiple, for example "15T".
            ntsd[ntsd.index.name + "_"] = (ntsd.index - ntsd.index[0]) // pd.Timedelta(
                ntsd.index.freqstr
            )
        except ValueError:
            ntsd[ntsd.index.name + "_"] = (ntsd.index - ntsd.index[0]) // pd.Timedelta(
                "1" + ntsd.index.freqstr
            )

    if x_pred_cols is None:
        nx_pred_cols = x_train_cols
    else:
        nx_pred_cols = x_pred_cols

    x_train_cols = tsutils.make_iloc(ntsd.columns, x_train_cols)
    y_train_col = tsutils.make_iloc(ntsd.columns, y_train_col)

    wtsd = ntsd.iloc[:, x_train_cols + y_train_col]

    # Train on 'any' dropna rows
    wtsddna = wtsd.dropna()
    # Train on last column
    y_train = wtsddna.iloc[:, -1].values
    # with all other columns
    x_train = wtsddna.iloc[:, :-1].values

    regr = _FUNCS[method]()
    regr.fit(x_train, y_train)

    if x_pred_cols is None:
        x_pred = x_train
    else:
        nx_pred_cols = tsutils.make_iloc(ntsd.columns, x_pred_cols)
        x_pred = ntsd.iloc[:, nx_pred_cols].dropna()
    y_pred = regr.predict(x_pred)

    if x_pred_cols is None:
        if method == "RANSAC":
            regr = regr.estimator_
        rdata = []
        rdata.append(["Coefficients", regr.coef_])
        rdata.append(["Intercept", regr.intercept_])
        rdata.append(["Mean squared error", mean_squared_error(y_train, y_pred)])
        rdata.append(["Coefficient of determination", r2_score(y_train, y_pred)])
        return rdata
    result = pd.DataFrame(y_pred, index=x_pred.index)
    result = result.reindex(index=wtsd.index)
    return tsutils.return_input(print_input, tsd, result)


if __name__ == "__init__":
    pass



# -*- coding: utf-8 -*-

import shlex
import subprocess
from unittest import TestCase

import pandas as pd
from pandas.testing import assert_frame_equal

import tstoolbox.tsutils as tsutils
from tstoolbox import tstoolbox

from . import capture


class TestDescribe(TestCase):
    def setUp(self):
        index = ["count", "mean", "std", "min", "25%", "50%", "75%", "max"]
        self.date_slice = pd.DataFrame(
            [
                1672.000000,
                836.905383,
                843.901292,
                0.000000,
                158.425000,
                578.800000,
                1253.450000,
                4902.000000,
            ],
            index=index,
            columns=["Area"],
        )
        self.date_slice.index.name = "Statistic"
        self.date_slice_cli = capture.capture(
            tsutils.printiso, self.date_slice, showindex="always"
        )
        self.date_slice.index.name = "UniqueID"

    def test_describe(self):
        """Test of describe API."""
        out = tstoolbox.describe(input_ts="tests/data_sunspot.csv")
        out.index.name = "UniqueID"
        assert_frame_equal(out, self.date_slice)

    def test_describe_cli(self):
        """Test of describe CLI."""
        args = 'tstoolbox describe --input_ts="tests/data_sunspot.csv"'
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        self.assertEqual(out, self.date_slice_cli)



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
import numpy as np
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

warnings.filterwarnings("ignore")


@mando.command("remove_trend", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def remove_trend_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    round_index=None,
    source_units=None,
    target_units=None,
    print_input=False,
    tablefmt="csv",
):
    """Remove a 'trend'.

    Subtracts from the data a linearly interpolated trend of the data.

    Parameters
    ----------
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${round_index}
    ${source_units}
    ${target_units}
    ${print_input}
    ${tablefmt}
    """
    tsutils.printiso(
        remove_trend(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            round_index=round_index,
            source_units=source_units,
            target_units=target_units,
            print_input=print_input,
        ),
        tablefmt=tablefmt,
    )


@tsutils.copy_doc(remove_trend_cli)
def remove_trend(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    round_index=None,
    source_units=None,
    target_units=None,
    print_input=False,
):
    """Remove a 'trend'."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    ntsd = tsd.copy()
    for col in tsd.columns:
        index = tsd.index.astype("l")
        index = index - index[0]
        lin = np.polyfit(index, tsd[col], 1)
        ntsd[col] = lin[0] * index + lin[1]
        ntsd[col] = tsd[col] - ntsd[col]
    return tsutils.return_input(print_input, tsd, ntsd, "remtrend")



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

warnings.filterwarnings("ignore")


@mando.command("pick", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def pick_cli(
    columns,
    input_ts="-",
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    clean=False,
    tablefmt="csv",
):
    """Will pick a column or list of columns from input [DEPRECATED].

    DEPRECATED: Effectively replaced by the "columns" keyword available
    in all other functions.

    Will be removed in a future version of `tstoolbox`.

    Can use column names or column numbers.  If using numbers, column
    number 1 is the first data column.

    Parameters
    ----------
    ${columns}
    ${input_ts}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        pick(
            columns,
            input_ts=input_ts,
            start_date=start_date,
            end_date=end_date,
            round_index=round_index,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            clean=clean,
        ),
        tablefmt=tablefmt,
    )


@tsutils.copy_doc(pick_cli)
def pick(
    columns,
    input_ts="-",
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    clean=False,
):
    """Will pick a column or list of columns from input."""
    warnings.warn(
        """
*
*   DEPRECATED in favor of using the "columns" keyword available in all
*   other functions.
*
*   Will be removed in a future version of `tstoolbox`.
*
"""
    )
    return tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

warnings.filterwarnings("ignore")


@mando.command("describe", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def describe_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    transpose=False,
    tablefmt="csv",
):
    """Print out statistics for the time-series.

    Parameters
    ----------
    transpose
        [optional, default is False, output format]

        If 'transpose' option is used, will transpose describe output.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${tablefmt}
    """
    tsutils.printiso(
        describe(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            transpose=transpose,
        ),
        showindex="always",
        tablefmt=tablefmt,
    )


@typic.al
@tsutils.copy_doc(describe_cli)
def describe(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    transpose: bool = False,
):
    """Print out statistics for the time-series."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        dropna=dropna,
        clean=clean,
    )
    if transpose is True:
        ntsd = tsd.describe().transpose()
    else:
        ntsd = tsd.describe()

    ntsd.index.name = "Statistic"
    return ntsd



# -*- coding: utf-8 -*-
import numpy as np

from . import centered_rms_dev, utils


def taylor_statistics(predicted, reference, field=""):
    """
    Calculate the statistics needed to create a Taylor diagram.

    Described in Taylor (2001) using the data provided in the predicted
    field (PREDICTED) and the reference field (REFERENCE).

    The statistics are returned in the STATS dictionary.

    If a dictionary is provided for PREDICTED or REFERENCE, then
    the name of the field must be supplied in FIELD.

    The function currently supports dictionaries, lists, and np.ndarray,
    types for the PREDICTED and REFERENCE variables.

    Input:
    PREDICTED : predicted field
    REFERENCE : reference field
    FIELD     : name of field to use in PREDICTED and REFERENCE dictionaries
                (optional)
    NORM      : logical flag specifying statistics are to be normalized
                with respect to standard deviation of reference field
                = True,  statistics are normalized
                = False, statistics are not normalized

    Output:
    STATS          : dictionary containing statistics
    STATS['ccoef'] : correlation coefficients (R)
    STATS['crmsd'] : centered root-mean-square (RMS) differences (E')
    STATS['sdev']  : standard deviations

    Each of these outputs are one-dimensional with the same length.
    First index corresponds to the reference series for the diagram.
    For example SDEV[1] is the standard deviation of the reference
    series (sigma_r) and SDEV[2:N] are the standard deviations of the
    other (predicted) series.

    Reference:

    Taylor, K. E. (2001), Summarizing multiple aspects of model
      performance in a single diagram, J. Geophys. Res., 106(D7),
      7183-7192, doi:10.1029/2000JD900719.

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 3, 2016
    """
    # Check for valid arguments
    if isinstance(predicted, dict):
        if field == "":
            raise ValueError("FIELD argument not supplied.")
        if field in predicted:
            p = predicted[field]
        else:
            raise ValueError("Field is not in PREDICTED dictionary: " + field)
    elif isinstance(predicted, list):
        p = np.array(predicted)
    elif isinstance(predicted, np.ndarray):
        p = predicted
    else:
        raise ValueError("PREDICTED argument must be a dictionary.")

    if isinstance(reference, dict):
        if field == "":
            raise ValueError("FIELD argument not supplied.")
        if field in reference:
            r = reference[field]
        else:
            raise ValueError("Field is not in REFERENCE dictionary: " + field)
    elif isinstance(reference, list):
        r = np.array(reference)
    elif isinstance(reference, np.ndarray):
        r = reference
    else:
        raise ValueError("REFERENCE argument must be a dictionary.")

    # Check that dimensions of predicted and reference fields match
    utils.check_arrays(predicted, reference)

    # Calculate correlation coefficient
    ccoef = np.corrcoef(p, r)
    ccoef = ccoef[0]

    # Calculate centered root-mean-square (RMS) difference (E')^2
    crmsd = [0.0, centered_rms_dev(p, r)]

    # Calculate standard deviation of predicted field w.r.t N (sigma_p)
    sdevp = np.std(p)

    # Calculate standard deviation of reference field w.r.t N (sigma_r)
    sdevr = np.std(r)
    sdev = [sdevr, sdevp]

    # Store statistics in a dictionary
    stats = {"ccoef": ccoef, "crmsd": crmsd, "sdev": sdev}
    return stats



# -*- coding: utf-8 -*-
import numpy as np

from . import centered_rms_dev, utils


def target_statistics(predicted, reference, field="", norm=False):
    """
    Calculate the statistics needed to create a target diagram.

    Described in Jolliff et al. (2009) using the data provided in the
    predicted field (PREDICTED) and the reference field (REFERENCE).

    The statistics are returned in the STATS dictionary.

    If a dictionary is provided for PREDICTED or REFERENCE, then
    the name of the field must be supplied in FIELD.

    The function currently supports dictionaries, lists, and np.ndarray,
    types for the PREDICTED and REFERENCE variables.

    Input:
    PREDICTED : predicted field
    REFERENCE : reference field
    FIELD     : name of field to use in PREDICTED and REFERENCE dictionaries
                (optional)
    NORM      : logical flag specifying statistics are to be normalized
                with respect to standard deviation of reference field
                = True,  statistics are normalized
                = False, statistics are not normalized

    Output:
    STATS          : dictionary containing statistics
    STATS['bias']  : bias (B)
    STATS['crmsd'] : centered root-mean-square (RMS) differences (E')
    STATS['rmsd']  : total RMS difference (RMSD)

    Each of these outputs are one-dimensional with the same length.

    Reference:

    Jolliff, J. K., J. C. Kindle, I. Shulman, B. Penta, M. Friedrichs,
      R. Helber, and R. Arnone (2009), Skill assessment for coupled
      biological/physical models of marine systems, J. Mar. Sys., 76(1-2),
      64-82, doi:10.1016/j.jmarsys.2008.05.014

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Nov 24, 2016
    """
    # Check for valid arguments
    if isinstance(predicted, dict):
        if field == "":
            raise ValueError("FIELD argument not supplied.")
        if field in predicted:
            p = predicted[field]
        else:
            raise ValueError("Field is not in PREDICTED dictionary: " + field)
    elif isinstance(predicted, list):
        p = np.array(predicted)
    elif isinstance(predicted, np.ndarray):
        p = predicted
    else:
        raise ValueError("PREDICTED argument must be a dictionary.")

    if isinstance(reference, dict):
        if field == "":
            raise ValueError("FIELD argument not supplied.")
        if field in reference:
            r = reference[field]
        else:
            raise ValueError("Field is not in REFERENCE dictionary: " + field)
    elif isinstance(reference, list):
        r = np.array(reference)
    elif isinstance(reference, np.ndarray):
        r = reference
    else:
        raise ValueError("REFERENCE argument must be a dictionary.")

    # Check that dimensions of predicted and reference fields match
    utils.check_arrays(p, r)

    # Calculate bias (B)
    bias = np.mean(p) - np.mean(r)

    # Calculate centered root-mean-square (RMS) difference (E')
    crmsd = centered_rms_dev(p, r)

    # Calculate RMS difference (RMSD)
    rmsd = np.sqrt(np.sum(np.square(np.subtract(p, r))) / float(p.size))

    # Normalize if requested
    if norm is True:
        sigma_ref = np.std(r)
        bias = bias / sigma_ref
        crmsd = crmsd / sigma_ref
        rmsd = rmsd / sigma_ref

    # Store statistics in a dictionary
    stats = {"bias": bias, "crmsd": crmsd, "rmsd": rmsd}
    if norm is True:
        stats["type"] = "normalized"
    else:
        stats["type"] = "unnormalized"

    return stats



# -*- coding: utf-8 -*-
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import pytest

from tstoolbox import tstoolbox

# Pull this in once.
df = tstoolbox.aggregate(
    agg_interval="D", clean=True, input_ts="tests/02234500_65_65.csv", skipna=False
)
# Pull this in once.
dfa = tstoolbox.aggregate(
    agg_interval="A", clean=True, input_ts="tests/02234500_65_65.csv", skipna=False
)


@pytest.mark.mpl_image_compare(tolerance=6)
def test_histogram():
    plt.close("all")
    return tstoolbox.plot(
        type="histogram",
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        sharex=False,
        plot_styles="classic",
    )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_heatmap():
#     plt.close("all")
#     return tstoolbox.plot(
#         type="heatmap",
#         columns=2,
#         clean=True,
#         input_ts=df,
#         ofilename=None,
#         plot_styles="classic",
#     )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_norm_xaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="norm_xaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_norm_yaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="norm_yaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_lognorm_xaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="lognorm_xaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_lognorm_yaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="lognorm_yaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_weibull_xaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="weibull_xaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_weibull_yaxis():
    plt.close("all")
    return tstoolbox.plot(
        type="weibull_yaxis",
        columns=2,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_kde_time():
#     plt.close("all")
#     return tstoolbox.plot(
#         type="kde_time",
#         columns=2,
#         clean=True,
#         input_ts="tests/02234500_65_65.csv",
#         ofilename=None,
#         plot_styles="classic",
#     )


# @pytest.mark.mpl_image_compare(tolerance=6)
# def test_kde_time_multiple_traces():
#     plt.close("all")
#     ndf = tstoolbox.read(["tests/daily.csv", "tests/02325000_flow.csv"])
#     return tstoolbox.plot(
#         type="kde_time",
#         columns=[2, 3],
#         clean=True,
#         input_ts=ndf,
#         ytitle="Flow",
#         ofilename=None,
#         plot_styles="classic",
#     )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_autocorrelation():
    plt.close("all")
    return tstoolbox.plot(
        type="autocorrelation",
        columns=2,
        input_ts=df,
        ofilename=None,
        plot_styles="classic",
    )



# -*- coding: utf-8 -*-
import numpy as np

from . import utils


def kge(simulated, observed):
    """
    Calculate the Kling-Gupta efficiency.

    Calculates the Kling-Gupta efficiency between two variables
    simulated and observed. The kge is calculated using the
    formula:

    KGE = 1 - sqrt((cc-1)**2 + (alpha-1)**2 + (beta-1)**2)

    where:
        cc = correlation coefficient between simulated and observed;
        alpha = std(simulated) / std(observed)
        beta = sum(simulated) / sum(observed)

    where s is the simulated values, o is the observed values, and
    N is the total number of values in s & o. Note that s & o must
    have the same number of values.

    Kling-Gupta efficiency can range from -infinity to 1. An efficiency of 1 (E
    = 1) corresponds to a perfect match of model to observed data.
    Essentially, the closer the model efficiency is to 1, the more accurate the
    model is.

    The efficiency coefficient is sensitive to extreme values and might yield
    sub-optimal results when the dataset contains large outliers in it.

    Kling-Gupta efficiency can be used to quantitatively describe the accuracy
    of model outputs. This method can be used to describe the predictive
    accuracy of other models as long as there is observed data to compare the
    model results to.

    Input:
    simulated : simulated values
    observed : observed values

    Output:
    kge : Kling-Gupta Efficiency
    """
    # Check that dimensions of simulated and observed fields match
    utils.check_arrays(simulated, observed)

    alpha = np.std(simulated) / np.std(observed)
    beta = np.sum(simulated) / np.sum(observed)
    cc = np.corrcoef(observed, simulated)[0, 1]

    # Calculate the kge
    kge = 1.0 - np.sqrt((cc - 1.0) ** 2 + (alpha - 1.0) ** 2 + (beta - 1.0) ** 2)

    return kge



# -*- coding: utf-8 -*-
"""A lag routine."""

from __future__ import absolute_import, print_function

from typing import List

import mando
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils


@mando.command("lag", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def lag_cli(
    lags,
    input_ts="-",
    print_input=False,
    start_date=None,
    end_date=None,
    columns=None,
    clean=False,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    skiprows=None,
    tablefmt="csv",
):
    """Create a series of lagged time-series.

    Parameters
    ----------
    lags : str, int, or list
        If an integer will calculate all lags up to and including the
        lag number.  If a list will calculate each lag in the list.  If
        a string must be a comma separated list of integers.
    ${print_input}
    ${input_ts}
    ${start_date}
    ${end_date}
    ${clean}
    ${skiprows}
    ${index_type}
    ${names}
    ${source_units}
    ${target_units}
    ${columns}
    ${tablefmt}
    """
    tsutils.printiso(
        lag(
            lags,
            input_ts=input_ts,
            print_input=print_input,
            start_date=start_date,
            end_date=end_date,
            columns=columns,
            clean=clean,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            skiprows=skiprows,
        ),
        tablefmt=tablefmt,
    )


@typic.constrained(ge=0)
class Lags(int):
    """'lags' constraint"""


@tsutils.transform_args(lags=tsutils.make_list)
@typic.al
@tsutils.copy_doc(lag_cli)
def lag(
    lags: List[Lags],
    input_ts="-",
    print_input=False,
    start_date=None,
    end_date=None,
    columns=None,
    clean=False,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    skiprows=None,
):
    """Create a series of lagged time-series."""
    tsd = tsutils.common_kwds(
        input_ts,
        dropna="all",
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )
    if len(lags) == 1:
        lags = lags[0]
    try:
        lags = list(range(1, lags + 1))
    except TypeError:
        pass

    if lags == 0:
        return tsd

    if print_input is True:
        ntsd = tsd.copy()
    else:
        ntsd = tsd

    ntsd = tsutils.asbestfreq(ntsd)

    cols = {}
    nlags = []
    for i in lags:
        for x in list(ntsd.columns):
            parts = x.split(":")
            parts[0] = "{}_{}".format(parts[0], i)
            cols.setdefault(x, []).append(":".join(parts))
            nlags.append(i)
    for k, v in cols.items():
        columns = v
        dfn = pd.DataFrame(data=None, columns=columns, index=ntsd.index)
        for c, i in zip(columns, lags):
            dfn[c] = ntsd[k].shift(periods=i)
        ntsd = pd.concat([ntsd, dfn], axis=1).reindex(ntsd.index)
    return tsutils.return_input(print_input, tsd, ntsd, "lag")



# -*- coding: utf-8 -*-
import warnings

import matplotlib.colors as clr
import matplotlib.pyplot as plt


def plot_pattern_diagram_markers(X, Y, option):
    """
    Plot color markers on a pattern diagram.

    Plots color markers on a target diagram according their (X,Y)
    locations. The symbols and colors are chosen automatically with a
    limit of 70 symbol & color combinations.

    The color bar is titled using the content of option['titleColorBar']
    (if non-empty string).

    INPUTS:
    x : x-coordinates of markers
    y : y-coordinates of markers
    z : z-coordinates of markers (used for color shading)
    option : dictionary containing option values. (Refer to
        GET_TARGET_DIAGRAM_OPTIONS function for more information.)
    option['axismax'] : maximum for the X & Y values. Used to limit
        maximum distance from origin to display markers
    option['markerlabel'] : labels for markers

    OUTPUTS:
    None

    Created on Nov 30, 2016

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com
    """
    # Set face color transparency
    alpha = option["alpha"]

    # Set marker size
    markerSize = option["markersize"]

    if option["markerlegend"] == "on":
        # Check that marker labels have been provided
        if option["markerlabel"] == "":
            raise ValueError("No marker labels provided.")

        # Plot markers of different color and shapes with labels
        # displayed in a legend

        # Define markers
        kind = ["+", "o", "x", "s", "d", "^", "v", "p", "h", "*"]
        colorm = ["b", "r", "g", "c", "m", "y", "k"]
        if len(X) > 70:
            _disp("You must introduce new markers to plot more than 70 cases.")
            _disp(
                "The " "marker" " character array need to be extended inside the code."
            )

        if len(X) <= len(kind):
            # Define markers with specified color
            marker = []
            markercolor = []
            for color in colorm:
                for symbol in kind:
                    marker.append(symbol + option["markercolor"])
                    rgba = clr.to_rgb(option["markercolor"]) + (alpha,)
                    markercolor.append(rgba)
        else:
            # Define markers and colors using predefined list
            marker = []
            for color in colorm:
                for symbol in kind:
                    marker.append(symbol + color)
                    rgba = clr.to_rgb(color) + (alpha,)
                    markercolor.append(rgba)

        # Plot markers at data points
        limit = option["axismax"]
        hp = ()
        markerlabel = []
        for i, xval in enumerate(X):
            if abs(X[i]) <= limit and abs(Y[i]) <= limit:
                h = plt.plot(
                    X[i],
                    Y[i],
                    marker[i],
                    markersize=markerSize,
                    markerfacecolor=markercolor[i],
                    markeredgecolor=marker[i][1],
                    markeredgewidth=2,
                )
                hp += tuple(h)
                markerlabel.append(option["markerlabel"][i])

        # Add legend
        if len(markerlabel) > 0:
            markerlabel = tuple(markerlabel)
            plt.legend(
                hp, markerlabel, loc="upper right", fontsize="medium", numpoints=1
            )
        else:
            warnings.warn("No markers within axis limit ranges.")

    else:
        # Plot markers as dots of a single color with accompanying labels
        # and no legend

        # Plot markers at data points
        limit = option["axismax"]
        rgba = clr.to_rgb(option["markercolor"]) + (alpha,)
        for i, xval in enumerate(X):
            if abs(X[i]) <= limit and abs(Y[i]) <= limit:
                # Plot marker
                plt.plot(
                    X[i],
                    Y[i],
                    ".",
                    markersize=markerSize,
                    markerfacecolor=rgba,
                    markeredgecolor=option["markercolor"],
                )

                # Check if marker labels provided
                if len(option["markerlabel"]) > 0:
                    # Label marker
                    xtextpos = X[i]  # ToDo: convert to double?
                    ytextpos = Y[i]
                    plt.text(
                        xtextpos,
                        ytextpos,
                        option["markerlabel"][i],
                        color=option["markerlabelcolor"],
                        verticalalignment="bottom",
                        horizontalalignment="right",
                        fontsize="medium",
                    )


def _disp(text):
    print(text)



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import warnings

import mando
import pandas as pd
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

warnings.filterwarnings("ignore")


@mando.command("stack", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def stack_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    clean=False,
    tablefmt="csv",
):
    """Return the stack of the input table.

    The stack command takes the standard table and
    converts to a three column table.

    From::

      Datetime,TS1,TS2,TS3
      2000-01-01 00:00:00,1.2,1018.2,0.0032
      2000-01-02 00:00:00,1.8,1453.1,0.0002
      2000-01-03 00:00:00,1.9,1683.1,-0.0004

    To::

      Datetime,Columns,Values
      2000-01-01,TS1,1.2
      2000-01-02,TS1,1.8
      2000-01-03,TS1,1.9
      2000-01-01,TS2,1018.2
      2000-01-02,TS2,1453.1
      2000-01-03,TS2,1683.1
      2000-01-01,TS3,0.0032
      2000-01-02,TS3,0.0002
      2000-01-03,TS3,-0.0004

    Parameters
    ----------
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        stack(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            round_index=round_index,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            source_units=source_units,
            target_units=target_units,
            clean=clean,
        ),
        tablefmt=tablefmt,
    )


@tsutils.copy_doc(stack_cli)
def stack(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    round_index=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    source_units=None,
    target_units=None,
    clean=False,
):
    """Return the stack of the input table."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    newtsd = pd.DataFrame(tsd.stack()).reset_index(1)
    newtsd.columns = ["Columns", "Values"]
    # newtsd = (
    #     newtsd.groupby("Columns")
    #     .apply(lambda d: d.sort_values("Columns"))
    #     .reset_index("Columns", drop=True)
    # )
    newtsd.sort_values("Columns", inplace=True)
    return newtsd



# -*- coding: utf-8 -*-
import numpy as np


def check_arrays(predicted, reference, pname="predicted", rname="reference"):
    """Generic check of input arrays."""
    pdims = predicted.shape
    rdims = reference.shape
    if not np.array_equal(pdims, rdims):
        raise ValueError(
            """
*
*   The {3} and {4} field dimensions do not match.
*       shape({3}) = {0}
*       shape({4}) = {1}
*       {3} type: {2}
*
""".format(
                pdims, rdims, type(predicted), pname, rname
            )
        )



# -*- coding: utf-8 -*-
from array import array

import numpy as np

from . import (
    get_target_diagram_axes,
    get_target_diagram_options,
    overlay_target_diagram_circles,
    plot_pattern_diagram_colorbar,
    plot_pattern_diagram_markers,
    plot_target_axes,
)


def target_diagram(*args, **kwargs):
    """
    Plot a target diagram from statistics of different series.

    target_diagram(bs,rmsds,rmsdxz,keyword=value)

    The first 3 arguments must be the inputs as described below followed by
    keywords in the format OPTION = value. An example call to the function
    would be:

    target_diagram(bs,rmsds,rmsdxz,markerdisplayed='marker')

    INPUTS:
    bs    : Bias (B) or Normalized Bias (B*). Plotted along y-axis
            as "Bias".
    rmsds : unbiased Root-Mean-Square Difference (RMSD') or normalized
            unbiased Root-Mean-Square Difference (RMSD*'). Plotted along
            x-axis as "uRMSD".
    rmsdxz : total Root-Mean-Square Difference (RMSD) or other quantities
            (if 'nonrmsdxz' == 'on'). Labeled on plot as "RMSD".

    OUTPUTS:
    None.

    LIST OF OPTIONS:
    For an exhaustive list of options to customize your diagram, call the
    function without arguments at a Python command line:
    % python
    >> from tstoolbox.skill_metrics.target_diagram import target_diagram
    >> target_diagram()

    Reference:

    Jolliff, J. K., J. C. Kindle, I. Shulman, B. Penta, M. Friedrichs,
      R. Helber, and R. Arnone (2009), Skill assessment for coupled
      biological/physical models of marine systems, J. Mar. Sys., 76(1-2),
      64-82, doi:10.1016/j.jmarsys.2008.05.014

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Nov 25, 2016
    """
    # Check for number of arguments
    nargin = len(args)
    bs, rmsds, rmsdxz = _get_target_diagram_arguments(*args)
    if nargin == 0:
        return

    # Get options
    option = get_target_diagram_options(**kwargs)

    #  Get axis values for plot
    axes = get_target_diagram_axes(rmsds, bs, option)

    # Plot axes for target diagram
    if option["overlay"] == "off":
        plot_target_axes(axes)

    # __ Overlay circles
    overlay_target_diagram_circles(option)

    # Plot data points
    lowcase = option["markerdisplayed"].lower()
    if lowcase == "marker":
        plot_pattern_diagram_markers(rmsds, bs, option)
    elif lowcase == "colorbar":
        plot_pattern_diagram_colorbar(rmsds, bs, rmsdxz, option)
    else:
        raise ValueError("Unrecognized option: " + option["markerdisplayed"])


def _get_target_diagram_arguments(*args):
    """
    Get arguments for target_diagram function.

    Retrieves the arguments supplied to the TARGET_DIAGRAM function as
    arguments and displays the optional arguments if none are supplied.
    Otherwise, tests the first 3 arguments are numeric quantities and
    returns their values.

    INPUTS:
    args : variable-length input argument list

    OUTPUTS:
    bs    : Bias (B) or Normalized Bias (B*). Plotted along y-axis
            as "Bias".
    rmsds : unbiased Root-Mean-Square Difference (RMSD') or normalized
            unbiased Root-Mean-Square Difference (RMSD*'). Plotted along
            x-axis as "uRMSD".
    rmsdxz : total Root-Mean-Square Difference (RMSD) or other quantities
            (if 'nonrmsdxz' == 'on'). Labeled on plot as "RMSD".
    """
    import numbers

    bs = []
    rmsds = []
    rmsdxz = []
    nargin = len(args)
    if nargin == 0:
        # Display options list
        _display_target_diagram_options()
        return bs, rmsds, rmsdxz
    if nargin != 3:
        raise ValueError("Must supply 3 arguments.")

    bs = args[0]
    rmsds = args[1]
    rmsdxz = args[2]

    # Test the above are numeric quantities
    if isinstance(bs, array):
        bs = np.array(bs)
    if isinstance(bs, numbers.Number):
        bs = np.array(bs, ndmin=1)
    if not isinstance(bs, np.ndarray):
        raise ValueError("Argument bs is not a numeric array")

    if isinstance(rmsds, array):
        rmsds = np.array(rmsds)
    if isinstance(rmsds, numbers.Number):
        rmsds = np.array(rmsds, ndmin=1)
    if not isinstance(rmsds, np.ndarray):
        raise ValueError("Argument rmsds is not a numeric array")

    if isinstance(rmsdxz, array):
        rmsdxz = np.array(rmsdxz)
    if isinstance(rmsdxz, numbers.Number):
        rmsdxz = np.array(rmsdxz, ndmin=1)
    if not isinstance(rmsdxz, np.ndarray):
        raise ValueError("Argument rmsdxz is not a numeric array")

    return bs, rmsds, rmsdxz


def _display_target_diagram_options():
    """Display available options for TARGET_DIAGRAM function."""
    _disp("General options:")
    _dispopt(
        "'overlay'",
        "'on' / 'off' (default): "
        + "Switch to overlay current statistics on target diagram. "
        + "\n\t\tOnly markers will be displayed.",
    )
    _dispopt(
        "'colormap'",
        "'on'/ 'off' (default): "
        + "Switch to map color shading of markers to colormap ('on')\n\t\t"
        + "or min to max range of rmsdxz values ('off').\n\t\t"
        + "Set to same value as option['nonrmsdxz'].",
    )
    _disp("")

    _disp("Marker options:")
    _dispopt(
        "'MarkerDisplayed'",
        "'marker' (default): Experiments are represented by individual symbols\n\t\t"
        + "'colorBar': Experiments are represented by a color described "
        + "in a colorbar",
    )
    _disp("OPTIONS when 'MarkerDisplayed' == 'marker'")
    _dispopt("'markerLabel'", "Labels for markers")
    _dispopt("'markerLabelColor'", "Marker label color (Default: black)")
    _dispopt("'markerColor'", "Marker color")
    _dispopt("'markerLegend'", "'on' / 'off' (default): Use legend for markers'")
    _dispopt("'markerSize'", "Marker size (Default: 10)")
    _disp("OPTIONS when 'MarkerDisplayed' == 'colorbar'")
    _dispopt(
        "'nonrmsdxz'",
        "'on'/ 'off' (default): "
        + "Values in rmsds do not correspond to total RMS Differences.\n\t\t"
        + "(Used to make range of rmsds values appear above color bar.)",
    )
    _dispopt("'titleColorBar'", "Title of the colorbar.")
    _disp("")

    _disp("Axes options:")
    _dispopt(
        "'ticks'", "define tick positions " + "(default is that used by axis function)"
    )
    _dispopt(
        "'xtickLabelPos'",
        "position of the tick labels " + "along the x-axis (empty by default)",
    )
    _dispopt(
        "'ytickLabelPos'",
        "position of the tick labels " + "along the y-axis (empty by default)",
    )
    _dispopt("'equalAxes'", "'on' (default) / 'off': Set axes to be equal")
    _dispopt("'limitAxis'", "Max for the Bias & uRMSD axis")
    _disp("")

    _disp("Diagram options:")
    _dispopt(
        "'alpha'",
        "Blending of symbol face color (0.0 transparent through 1.0 opaque)"
        + "\n\t\t"
        + "(Default: 1.0)",
    )
    _dispopt("'axismax'", "Maximum for the Bias & uRMSD axis")
    _dispopt(
        "'circles'",
        "define the radii of circles to draw "
        + "(default of (maximum rmsds)*[.7 1], [.7 1] when normalized diagram)",
    )
    _dispopt(
        "'circleLineSpec'",
        "Circle line specification (default " + "dashed black, '--k')",
    )
    _dispopt("'circleLineWidth'", "Circle line width")
    _dispopt("'obsUncertainty'", "Observational Uncertainty (default of 0)")
    _dispopt("'normalized'", "'on' / 'off' (default): normalized target diagram")


def _disp(text):
    print(text)


def _dispopt(optname, optval):
    """
    Display option name and values.

    This is a support function for the DISPLAY_TARGET_DIAGRAM_OPTIONS function.
    It displays the option name OPTNAME on a line by itself followed by its
    value OPTVAL on the following line.
    """
    _disp("\t%s" % optname)
    _disp("\t\t%s" % optval)



# -*- coding: utf-8 -*-
import numpy as np


def check_taylor_stats(stds, crmsds, cors, threshold=0.01):
    """
    Check that input statistics satisfy Taylor diagram relation to <1%.

    Function terminates with an error if not satisfied. The threshold is
    the ratio of the difference between the statistical metrics and the
    centered root mean square difference:

     abs(crmsds^2 - (stds^2 + stds(1)^2 - 2*stds*stds(1)*cors))/crmsds^2

    Note that the first element of the statistics vectors must contain
    the value for the reference field.

    INPUTS:
    stds      : Standard deviations
    crmsds    : Centered Root Mean Square Difference
    cors      : Correlation
    threshold : limit for acceptance, e.g. 0.1 for 10% (default 0.01)

    OUTPUTS:
    None.

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 3, 2016
    """
    if threshold < 1e-7:
        ValueError("threshold value must be positive: " + str(threshold))

    diff = np.square(crmsds[1:]) - (
        np.square(stds[1:])
        + np.square(stds[0])
        - 2.0 * stds[0] * np.multiply(stds[1:], cors[1:])
    )
    diff = np.abs(np.divide(diff, np.square(crmsds[1:])))
    index = np.where(diff > threshold)
    if len(index) > 0:
        ii = np.where(diff != 0)
        if len(ii) == len(diff):
            ValueError(
                """
*
*   Incompatible data
*
*   You must have:
*       crmsds - sqrt(stds.^2 + stds(1)^2 - 2*stds*stds(1).*cors) = 0
*
"""
            )
        else:
            ValueError(
                """
*
*   Incompatible data indices: {}
*
*   You must have:
*       crmsds - sqrt(stds.^2 + stds(1)^2 - 2*stds*stds(1).*cors) = 0
*
""".format(
                    ii
                )
            )

    return diff



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

from typing import Optional

import mando
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


@mando.command("pct_change", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def pct_change_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    periods=1,
    fill_method="pad",
    limit=None,
    freq=None,
    print_input=False,
    round_index=None,
    source_units=None,
    target_units=None,
    float_format="g",
    tablefmt="csv",
):
    """Return the percent change between times.

    Parameters
    ----------
    periods : int
        [optional, default is 1]

        The number of intervals to calculate percent change across.
    fill_method : str
        [optional, defaults to 'pad']

        Fill method for NA.  Defaults to 'pad'.
    limit
        [optional, defaults to None]

        Is the minimum number of consecutive NA values where no more filling
        will be made.
    freq : str
        [optional, defaults to None]

        A pandas time offset string to represent the interval.

        ${pandas_offset_codes}
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${source_units}
    ${target_units}
    ${print_input}
    ${float_format}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        pct_change(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            periods=periods,
            fill_method=fill_method,
            limit=limit,
            freq=freq,
            print_input=print_input,
            round_index=round_index,
            source_units=source_units,
            target_units=target_units,
        ),
        float_format=float_format,
        tablefmt=tablefmt,
    )


@typic.al
@tsutils.copy_doc(pct_change_cli)
def pct_change(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    periods: tsutils.IntGreaterEqualToOne = 1,
    fill_method: Literal["backfill", "bfill", "ffill", "pad"] = "pad",
    limit: Optional[tsutils.IntGreaterEqualToZero] = None,
    freq: str = None,
    print_input=False,
    round_index=None,
    source_units=None,
    target_units=None,
    float_format="g",
):
    """Return the percent change between times."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    # Trying to save some memory
    if print_input:
        otsd = tsd.copy()
    else:
        otsd = pd.DataFrame()

    return tsutils.return_input(
        print_input,
        otsd,
        tsd.pct_change(
            periods=periods, fill_method=fill_method, limit=limit, freq=freq
        ),
        "pctchange",
    )



# -*- coding: utf-8 -*-

from unittest import TestCase

import pandas as pd
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class TestRead(TestCase):
    def setUp(self):
        """Prepare in-memory files data_stacked.csv and data_unstacked.csv."""
        self.stacked = pd.read_csv(
            "tests/data_stacked.csv", index_col=0, parse_dates=True
        )
        self.stacked.index.name = "Datetime"

        self.stacked_1 = pd.read_csv(
            "tests/data_stacked_1.csv", index_col=0, parse_dates=True
        ).convert_dtypes()
        self.stacked_1.index.name = "Datetime"

        self.unstacked = pd.read_csv(
            "tests/data_unstacked.csv", index_col=0, parse_dates=True
        )
        self.unstacked.rename(columns=lambda x: x.strip("'\" "))
        self.unstacked.index.name = "Datetime"

    def test_stack(self):
        """Stack the data_unstacked.csv file.

        Compare against the in-memory version of the data_stacked.csv file.
        """
        out = tstoolbox.stack(input_ts="tests/data_unstacked.csv").convert_dtypes()
        assert_frame_equal(out, self.stacked_1)

    def test_unstack(self):
        """Unstack the data_stacked.csv file.

        Compare against the in-memory version of the data_unstacked.csv file.
        """
        out = tstoolbox.unstack("Columns", input_ts="tests/data_stacked.csv")
        assert_frame_equal(
            tsutils.memory_optimize(out), tsutils.memory_optimize(self.unstacked)
        )



# -*- coding: utf-8 -*-

from __future__ import print_function

import shlex
import subprocess
from unittest import TestCase

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils

from . import capture


class TestEquation(TestCase):
    def setUp(self):
        dindex = pd.date_range("2000-01-01T00:00:00", periods=2, freq="D")
        ts1 = pd.Series([4.5, 4.6], index=dindex)
        ts2 = pd.Series([20.0, 20.4], index=dindex)
        self.equation = pd.DataFrame(
            {"Value": ts1, "Value::equation0_": ts2}, index=dindex, dtype="float64"
        )
        self.equation.index.name = "Datetime"
        self.equation = tsutils.memory_optimize(self.equation)

        self.equation_cli = capture.capture(tsutils.printiso, self.equation)

        dindex = pd.date_range("2000-01-01T00:00:00", periods=6, freq="D")
        ts1 = [4.50, 4.60, 4.70, 4.60, 4.50, 4.40]
        ts2 = [45.60, 90.50, 34.20, 23.10, 7.20, 4.30]
        ts3 = [
            -4.00389911875,
            -5.61841686236,
            -5.87322876716,
            -4.86614401542,
            -2.6934178416,
            -4.60800663972,
        ]
        self.equation_multiple_cols_01 = pd.DataFrame(
            {"Value": ts1, "Value1": ts2, "_::equation0_": ts3},
            index=dindex,
            dtype="float64",
        )
        self.equation_multiple_cols_01.index.name = "Datetime"
        self.equation_multiple_cols_01 = tsutils.memory_optimize(
            self.equation_multiple_cols_01
        )

        self.equation_result = pd.DataFrame(
            {"_::equation0_": np.array(ts2) * 10}, index=dindex, dtype="Float64"
        )
        self.equation_result = tsutils.memory_optimize(self.equation_result)

        self.equation_multiple_cols_01_cli = capture.capture(
            tsutils.printiso, self.equation_multiple_cols_01
        )

        ts3 = [50.1, 95.1, 38.9, 27.7, 11.7, 8.7]
        self.equation_multiple_cols_02 = pd.DataFrame(
            {"Value": ts1, "Value1": ts2, "_::equation0_": ts3},
            index=dindex,
            dtype="float64",
        )
        self.equation_multiple_cols_02.index.name = "Datetime"
        self.equation_multiple_cols_02 = tsutils.memory_optimize(
            self.equation_multiple_cols_02
        )

        self.equation_multiple_cols_02_cli = capture.capture(
            tsutils.printiso, self.equation_multiple_cols_02
        )

        ts3 = [0, 97.92, 41.66, 30.52, 14.46, 0]
        ts3[0] = np.nan
        ts3[-1] = np.nan
        self.equation_multiple_cols_03 = pd.DataFrame(
            {"Value": ts1, "Value1": ts2, "_::equation0_": ts3},
            index=dindex,
            dtype="float64",
        )
        self.equation_multiple_cols_03.index.name = "Datetime"
        self.equation_multiple_cols_03 = tsutils.memory_optimize(
            self.equation_multiple_cols_03
        )

        self.equation_multiple_cols_03_cli = capture.capture(
            tsutils.printiso, self.equation_multiple_cols_03, float_format=".2f"
        )

        dindex = pd.date_range("2011-01-01T00:00:00", periods=48, freq="H")
        ts1 = [2] * 48
        ts2 = [5.2] * 48
        ts2[0] = np.nan
        ts2[-1] = np.nan
        self.equation_multiple_cols_04 = pd.DataFrame(
            {"Value": ts1, "Value::equation0_": ts2}, index=dindex
        ).infer_objects()
        self.equation_multiple_cols_04.index.name = "Datetime"
        self.equation_multiple_cols_04 = tsutils.memory_optimize(
            self.equation_multiple_cols_04
        )

        self.equation_multiple_cols_04_cli = capture.capture(
            tsutils.printiso, self.equation_multiple_cols_04
        )

    def test_equation(self):
        """Test equation API."""
        out = tstoolbox.equation(
            "x*4 + 2", input_ts="tests/data_simple.csv", print_input=True
        )
        assert_frame_equal(out, self.equation, check_column_type=False)

    def test_equation_multiple_cols_01(self):
        """Test of equation API with multiple columns and numpy functions."""
        out = tstoolbox.equation(
            "sin(x1)*4 + cos(x2)*2",
            input_ts="tests/data_multiple_cols.csv",
            print_input=True,
        )
        assert_frame_equal(out, self.equation_multiple_cols_01, check_column_type=False)

    def test_equation_multiple_cols_02(self):
        """Test API with mulitple columns."""
        out = tstoolbox.equation(
            "x1 + x2", input_ts="tests/data_multiple_cols.csv", print_input=True
        )
        assert_frame_equal(out, self.equation_multiple_cols_02, check_column_type=False)

    def test_equation_multiple_cols_03(self):
        """Test API with multiple columns and different times records."""
        out = tstoolbox.equation(
            "x1[t] + 0.6*maximum(x1[t-1], x1[t+1]) + x2",
            input_ts="tests/data_multiple_cols.csv",
            print_input=True,
        )
        assert_frame_equal(out, self.equation_multiple_cols_03, check_column_type=False)

    def test_equation_multiple_cols_04(self):
        """Test of equation API using different time records."""
        out = tstoolbox.equation(
            "x[t] + 0.6*maximum(x[t-1], x[t+1]) + x[t]",
            input_ts="tests/data_flat.csv",
            print_input=True,
        )
        self.maxDiff = None
        assert_frame_equal(out, self.equation_multiple_cols_04, check_column_type=False)

    def test_equation_multiple_cols_05(self):
        """Test of equation API using different time records.

        Almost same as test_equation_multiple_cols_04
        """
        out = tstoolbox.equation(
            "x + 0.6*maximum(x[t-1], x[t+1]) + x[t]",
            input_ts="tests/data_flat.csv",
            print_input=True,
        )
        self.maxDiff = None
        assert_frame_equal(out, self.equation_multiple_cols_04, check_column_type=False)

    def test_equation_cols_over_nine(self):
        """Test of using equation API with columns over 9."""
        input_ts = tstoolbox.read(
            "tests/data_multiple_cols.csv tests/data_multiple_cols.csv tests/data_multiple_cols.csv tests/data_multiple_cols.csv tests/data_multiple_cols.csv",
            append="columns",
        )
        out = tstoolbox.equation("x10*10", input_ts=input_ts)
        assert_frame_equal(out, self.equation_result, check_column_type=False)

    def test_equation_cli(self):
        """Test of equation CLI."""
        args = (
            'tstoolbox equation "x*4 + 2" '
            '--input_ts="tests/data_simple.csv"  '
            "--print_input=True"
        )
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        self.maxDiff = None
        self.assertEqual(out, self.equation_cli)

    def test_equation_multiple_cols_01_cli(self):
        """Test of equation API with multiple columns."""
        args = (
            'tstoolbox equation "sin(x1)*4 + cos(x2)*2" '
            '--input_ts="tests/data_multiple_cols.csv" '
            "--print_input=True"
        )
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        out = tsutils.read_iso_ts(out)
        assert_frame_equal(
            out, tstoolbox.date_slice(self.equation_multiple_cols_01_cli)
        )

    def test_equation_multiple_cols_02_cli(self):
        """Test of equation CLI with multiple columns."""
        args = (
            'tstoolbox equation "x1 + x2" '
            '--input_ts="tests/data_multiple_cols.csv" '
            "--print_input=True"
        )
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        out = tsutils.read_iso_ts(out)
        assert_frame_equal(out, self.equation_multiple_cols_02)

    def test_equation_multiple_cols_03_cli(self):
        """Test of equation CLI with multiple columns and time records."""
        args = (
            "tstoolbox equation "
            '"x1[t] + 0.6*maximum(x1[t-1], x1[t+1]) + x2" '
            '--input_ts="tests/data_multiple_cols.csv" '
            "--print_input=True "
            '--float_format=".2f"'
        )
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        out = tsutils.read_iso_ts(out)
        assert_frame_equal(out, self.equation_multiple_cols_03)

    def test_equation_multiple_cols_04_cli(self):
        """Test of equation CLI with multiple time records."""
        args = (
            "tstoolbox equation "
            '"x[t] + 0.6*maximum(x[t-1], x[t+1]) + x[t]" '
            '--input_ts="tests/data_flat.csv" '
            "--print_input=True"
        )
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        self.maxDiff = None
        self.assertEqual(out, self.equation_multiple_cols_04_cli)

    def test_equation_multiple_cols_05_cli(self):
        """Test of equation with multiple time records.

        Almost same as test_equation_multiple_cols_04_cli.
        """
        args = (
            "tstoolbox equation "
            '"x + 0.6*maximum(x[t-1], x[t+1]) + x[t]" '
            '--input_ts="tests/data_flat.csv" '
            "--print_input=True"
        )
        args = shlex.split(args)
        out = subprocess.Popen(
            args, stdout=subprocess.PIPE, stdin=subprocess.PIPE
        ).communicate()[0]
        assert_frame_equal(
            tstoolbox.date_slice(input_ts=out),
            tstoolbox.date_slice(self.equation_multiple_cols_04_cli),
        )



# -*- coding: utf-8 -*-

from unittest import TestCase

from tstoolbox import tstoolbox


class Testgof(TestCase):
    def setUp(self):
        self.df = tstoolbox.read(["tests/data_sunspot.csv", "tests/data_sunspot.csv"])
        self.gof = [
            ["Mean error or bias", 0.0],
            ["Root-mean-square Deviation/Error (RMSD)", 0.0],
            ["Centered RMSD (CRMSD)", 0.0],
            ["Pearson coefficient of correlation (r)", 1.0],
            ["Skill score (Murphy)", 1.0],
            ["Nash-Sutcliffe Efficiency", 1.0],
            ["Brier's Score", 0.0],
            ["Common count observed and simulated", 10],
            ["Count of NaNs", None, 0, 0],
        ]

    def test_gof(self):
        """Test of gof API."""
        out = tstoolbox.gof(
            obs_col=self.df.iloc[10:20, 0],
            sim_col=self.df.iloc[10:20, 1],
            stats=["me", "rmsd", "crmsd", "corrcoef", "murphyss", "nse", "brierss"],
        )
        assert out == self.gof



# -*- coding: utf-8 -*-

import os
import shlex
import subprocess
import sys

from setuptools import find_packages, setup

# temporarily redirect config directory to prevent matplotlib importing
# testing that for writeable directory which results in sandbox error in
# certain easy_install versions
os.environ["MPLCONFIGDIR"] = "."

pkg_name = "tstoolbox"

version = open("VERSION").readline().strip()

if sys.argv[-1] == "publish":
    subprocess.run(shlex.split("cleanpy ."))
    subprocess.run(shlex.split("python setup.py sdist"))
    subprocess.run(shlex.split(f"twine upload dist/{pkg_name}-{version}.tar.gz"))
    sys.exit()

README = open("README.rst").read()

install_requires = [
    # List your project dependencies here.
    # For more details, see:
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
    "dateparser",
    "tabulate",
    "docutils",
    "mando >= 0.4",
    "rst2ansi >= 0.1.5",
    "python-dateutil >= 2.1",
    "numpy",
    "scipy",
    "pandas",
    "pint-pandas",
    "matplotlib",
    "xlsxwriter",
    "scikit-learn",
    "statsmodels",
    "setuptools",
    "typical>=2.6.0",
    "typing-extensions",
    "pyaf",
    "xlrd >= 1.0.0",
    "openpyxl",
    "plottoolbox",
    "HydroErr",
]

extras_require = {
    "dev": [
        "black",
        "cleanpy",
        "twine",
        "pytest",
        "coverage[toml]",
        "flake8",
        "pytest-cov",
        "pytest-mpl",
        "pre-commit",
        "black-nbconvert",
        "blacken-docs",
        "velin",
        "isort",
        "pyroma",
        "pyupgrade",
        "commitizen",
    ]
}

setup(
    name=pkg_name,
    version=version,
    description="Command line script to manipulate time series files.",
    long_description=README,
    classifiers=[
        # Get strings from
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],  # strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords="time_series",
    author="Tim Cera, PE",
    author_email="tim@cerazone.net",
    url=f"http://timcera.bitbucket.io/{pkg_name}/docs/index.html",
    license="BSD",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"SciencePlots": ["*.mplstyle"]},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={"console_scripts": [f"{pkg_name}={pkg_name}.{pkg_name}:main"]},
    test_suite="tests",
    python_requires=">=3.7.1",
)



# -*- coding: utf-8 -*-
from . import brier_score, utils


def skill_score_brier(forecast, reference, observed):
    """
    Calculate Brier skill score (BSS) between two variables.

    Calculate non-dimensional skill score (SS) between two variables using
    definition of Murphy (1988)

    Calculates the non-dimensional skill score (SS) difference between two
    probabilities FORECAST and REFERENCE relative to OBSERVED outcomes. The
    skill score is calculated using the formula:

    BSS = 1 - BS/BSref

    where BS & BSref are the Brier scores for the forecast and reference
    probability of event occurrence

    BS    = sum_(n=1)^N (f_n - o_n)^2/N
    BSref = sum_(n=1)^N (r_n - o_n)^2/N

    where f is the forecast probabilities, o is the observed probabilities
    (0 or 1), and N is the total number of values in f & o. Note that f, r,
    & o must have the same number of values.

    Input:
    FORECAST  : forecast probabilities
    REFERENCE : reference probabilities
    OBSERVED  : observed probabilities

    Output:
    BSS : Brier skill score

    Reference:
    Glenn W. Brier, 1950: Verification of forecasts expressed in terms
    of probabilities. Mon. We. Rev., 78, 1-23.

    D. S. Wilks, 1995: Statistical Methods in the Atmospheric Sciences.
    Cambridge Press. 547 pp.

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com

    Created on Dec 15, 2016
    """
    # Check that dimensions of forecast, reference, and observed fields match
    utils.check_arrays(forecast, reference)
    utils.check_arrays(forecast, observed)

    # Calculate skill score
    bss = 1 - brier_score(forecast, observed) / brier_score(reference, observed)
    return bss



# -*- coding: utf-8 -*-
def check_on_off(value):
    """
    Check whether variable contains a value of 'on', 'off', True, or False.

    Returns an error if neither for the first two, and sets True to 'on',
    and False to 'off'. The 'on' and 'off' can be provided in any combination
    of upper and lower case letters.

    INPUTS:
    value : string or boolean to check

    OUTPUTS:
    None.

    Author: Peter A. Rochford
        Symplectic, LLC
        www.thesymplectic.com
        prochford@thesymplectic.com
    """
    if isinstance(value, str):
        lowcase = value.lower()
        if lowcase == "off":
            return lowcase
        if lowcase == "on":
            return lowcase
        raise ValueError("Invalid value: " + str(value))
    if isinstance(value, bool):
        if not value:
            value = "off"
        elif value:
            value = "on"
    else:
        raise ValueError("Invalid value: " + str(value))

    return value



# -*- coding: utf-8 -*-
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import pytest

from tstoolbox import tstoolbox

# Pull this in once.
df = tstoolbox.aggregate(
    agg_interval="D", clean=True, input_ts="tests/02234500_65_65.csv"
)
# Pull this in once.
dfa = tstoolbox.aggregate(
    agg_interval="A", clean=True, input_ts="tests/02234500_65_65.csv"
)


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_plot():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=1,
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_multiple_traces_plot():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=[2, 3],
        style="b-,r*",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_multiple_traces_style_plot():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=[2, 3],
        style="b-,r  ",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_multiple_traces_new_style_plot():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=[2, 3],
        markerstyles=" ,*",
        linestyles="-, ",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_time_markers():
    plt.close("all")
    return tstoolbox.plot(
        type="time",
        columns=[2, 3],
        linestyles=" ",
        markerstyles="auto",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        clean=True,
        input_ts="tests/02234500_65_65.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy_multiple_traces():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        columns=[2, 3, 3, 2],
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy_multiple_traces_logy():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        columns=[2, 3, 3, 2],
        yaxis="log",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy_multiple_traces_logx():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        columns=[2, 3, 3, 2],
        xaxis="log",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )


@pytest.mark.mpl_image_compare(tolerance=6)
def test_xy_multiple_traces_markers():
    plt.close("all")
    return tstoolbox.plot(
        type="xy",
        columns=[2, 3, 3, 2],
        linestyles=" ",
        markerstyles="auto",
        input_ts="tests/data_daily_sample.csv",
        ofilename=None,
        plot_styles="classic",
    )



# -*- coding: utf-8 -*-
"""Collection of functions for the manipulation of time series."""

from __future__ import absolute_import, division, print_function

import mando
import pandas as pd
import typic
from mando.rst_text_formatter import RSTHelpFormatter

from .. import tsutils

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


@mando.command("normalization", formatter_class=RSTHelpFormatter, doctype="numpy")
@tsutils.doc(tsutils.docstrings)
def normalization_cli(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    mode="minmax",
    min_limit=0,
    max_limit=1,
    pct_rank_method="average",
    print_input=False,
    round_index=None,
    source_units=None,
    target_units=None,
    float_format="g",
    tablefmt="csv",
    with_centering=True,
    with_scaling=True,
    quantile_range=(0.25, 0.75),
):
    """Return the normalization of the time series.

    This scales the time-series.

    Parameters
    ----------
    mode : str
        [optional, default is 'minmax']

        minmax
            min_limit +
            (X-Xmin)/(Xmax-Xmin)*(max_limit-min_limit)

        zscore
            (X-mean(X))/stddev(X)

        pct_rank
            rank(X)*100/N

        maxabs
            Scale by absolute value between -1 and 1.

        normal
            Scale to unit normal.

        robust
            Robust scale to ranked quantile ranges.
    min_limit : float
        [optional, defaults to 0, used for mode=minmax]

        Defines the minimum limit of the minmax normalization.
    max_limit : float
        [optional, defaults to 1, used for mode=minmax]

        Defines the maximum limit of the minmax normalization.
    pct_rank_method : str
        [optional, defaults to 'average']

        Defines how tied ranks are broken.  Can be 'average', 'min', 'max',
        'first', 'dense'.
    with_centering : bool
        [optional, defaults to True, used when mode=robust]

        If True, center the data before scaling.
    with_scaling : bool
        [optional, defaults to True, used when mode=robust]

        If True, scale the data to interquartile range.
    quantile_range : tuple
        [optional, defaults to (0.25, 0.75)
        (q_min, q_max), 0.0 < q_min < q_max < 100.0]

        Quantile range used to calculate scale.
    ${input_ts}
    ${columns}
    ${start_date}
    ${end_date}
    ${dropna}
    ${skiprows}
    ${index_type}
    ${names}
    ${clean}
    ${print_input}
    ${float_format}
    ${source_units}
    ${target_units}
    ${round_index}
    ${tablefmt}
    """
    tsutils.printiso(
        normalization(
            input_ts=input_ts,
            columns=columns,
            start_date=start_date,
            end_date=end_date,
            dropna=dropna,
            skiprows=skiprows,
            index_type=index_type,
            names=names,
            clean=clean,
            mode=mode,
            min_limit=min_limit,
            max_limit=max_limit,
            pct_rank_method=pct_rank_method,
            print_input=print_input,
            round_index=round_index,
            source_units=source_units,
            target_units=target_units,
            with_centering=with_centering,
            with_scaling=with_scaling,
            quantile_range=quantile_range,
        ),
        float_format=float_format,
        tablefmt=tablefmt,
    )


@typic.al
@tsutils.copy_doc(normalization_cli)
def normalization(
    input_ts="-",
    columns=None,
    start_date=None,
    end_date=None,
    dropna="no",
    skiprows=None,
    index_type="datetime",
    names=None,
    clean=False,
    mode: Literal[
        "minmax", "zscore", "pct_rank", "maxabs", "normal", "robust"
    ] = "minmax",
    min_limit: float = 0,
    max_limit: float = 1,
    pct_rank_method: Literal["average", "min", "max", "first", "dense"] = "average",
    print_input=False,
    round_index=None,
    source_units=None,
    target_units=None,
    with_centering=True,
    with_scaling=True,
    quantile_range=(0.25, 0.75),
):
    """Return the normalization of the time series."""
    tsd = tsutils.common_kwds(
        input_ts,
        skiprows=skiprows,
        names=names,
        index_type=index_type,
        start_date=start_date,
        end_date=end_date,
        pick=columns,
        round_index=round_index,
        dropna=dropna,
        source_units=source_units,
        target_units=target_units,
        clean=clean,
    )

    # Trying to save some memory
    if print_input:
        otsd = tsd.copy()
    else:
        otsd = pd.DataFrame()

    if mode == "minmax":
        tsd = min_limit + (tsd - tsd.min()) / (tsd.max() - tsd.min()) * (
            max_limit - min_limit
        )
    elif mode == "zscore":
        tsd = (tsd - tsd.mean()) / tsd.std()
    elif mode == "pct_rank":
        tsd = tsd.rank(method=pct_rank_method, pct=True)
    elif mode == "maxabs":
        from sklearn.preprocessing import MaxAbsScaler

        tsd.loc[:, :] = MaxAbsScaler().fit_transform(tsd)
    elif mode == "normal":
        from sklearn.preprocessing import Normalizer

        tsd.loc[:, :] = Normalizer().fit_transform(tsd)
    elif mode == "robust":
        from sklearn.preprocessing import RobustScaler

        tsd.loc[:, :] = RobustScaler(
            with_centering=with_centering,
            with_scaling=with_scaling,
            quantile_range=quantile_range,
        ).fit_transform(tsd)
    tsd = tsutils.memory_optimize(tsd)
    return tsutils.return_input(print_input, otsd, tsd, "{}".format(mode))



# -*- coding: utf-8 -*-

from unittest import TestCase

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from tstoolbox import tstoolbox, tsutils


class Testconvert_index(TestCase):
    def setUp(self):
        self.read_direct = pd.read_csv(
            "tests/data_gainesville_daily_precip_index.csv", index_col=0, header=0
        )
        self.read_direct = tsutils.memory_optimize(self.read_direct)

    def test_convert_index_from_UTC(self):
        out = tstoolbox.convert_index(
            "number",
            input_ts="tests/data_gainesville_daily_precip.csv",
            interval="D",
            epoch="1979-12-30",
        )
        out.index.name = "1979-12-30_date"
        out.index = pd.Int64Index(out.index)
        assert_frame_equal(out, self.read_direct)


@pytest.mark.parametrize(
    "epoch,interval,expected",
    [
        ("1970-01-01", "D", 3653),
        ("unix", "D", 3653),
        ("unix", "s", 3653 * 24 * 60 * 60),
        ("1970-01-01", "H", 3653 * 24),
        ("julian", "D", 2444240.5),
        ("julian", "h", 2444240.5 * 24),
        ("reduced", "D", 2444240.5 - 2400000),
        ("modified", "D", 2444240.5 - 2400000.5),
        ("truncated", "D", np.floor(2444240.5 - 2440000.5)),
        ("dublin", "D", 2444240.5 - 2415020),
        ("cnes", "D", 2444240.5 - 2433282.5),
        ("ccsds", "D", 2444240.5 - 2436204.5),
        ("lop", "D", 2444240.5 - 2448622.5),
        ("lilian", "D", np.floor(2444240.5 - 2299159.5)),
        ("rata_die", "D", np.floor(2444240.5 - 1721424.5)),
        ("mars_sol", "D", (2444240.5 - 2405522) / 1.02749),
        ("julian", None, 2444240.5),
    ],
)
def test_epoch_interval(epoch, interval, expected):
    if interval is not None and (
        (epoch == "unix" and interval != "s") or (epoch == "julian" and interval != "D")
    ):
        with pytest.warns(UserWarning, match="Typically the "):
            out = tstoolbox.convert_index(
                "number",
                input_ts="tests/data_gainesville_daily_precip.csv",
                interval=interval,
                epoch=epoch,
            )
    else:
        out = tstoolbox.convert_index(
            "number",
            input_ts="tests/data_gainesville_daily_precip.csv",
            interval=interval,
            epoch=epoch,
        )
    assert out.index[1] == expected


def test_raises():
    with pytest.raises(ValueError):
        _ = tstoolbox.convert_index(
            "zebra",
            input_ts="tests/data_gainesville_daily_precip.csv",
            interval="D",
            epoch="julian",
        )
